#include <exec/types.h>
#include <exec/memory.h>
#include <exec/ports.h>
#include <clib/alib_protos.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef __MAXON__
#include <linkerfunc.h>
#include <pragma/exec_lib.h>
#include <pragma/dos_lib.h>
#else
#include <proto/exec.h>
#include <proto/dos_lib.h>
#endif

#include "doom.h"
#include "DoomAttackNet.h"
#include "ComPort.h"

/*=====================*/

doomdata_t **netbuffer;		// pointer to variable!!!
doomcom_t *doomcom;

char	**myargv;
int	myargc;

static void (*I_Error)(char *error, ...);
static int (*M_CheckParm)(char *check);

/*=====================*/

static struct ComPort *comport;
static struct ComMsg sendmsg,*getmsg;
static struct ComMsg fastmsg[NUM_FASTMSG];

static struct List *sendport[MAX_PLAYERS+1];
static WORD portnode[MAX_PLAYERS+1];

WORD playerid,consoleplayer;

/*=====================*/


void DAN_Init(register __a0 struct DANInitialization *daninit)
{
	struct DANInitialization *init=daninit;
		
#ifdef __MAXON__
	InitModules();
#endif

	// link function pointers to DoomAttack routines

	I_Error=init->I_Error;
	M_CheckParm=init->M_CheckParm;
	
	// setups vars
				
	netbuffer	= init->netbuffer;
	doomcom		= init->doomcom;
	myargv		= init->myargv;
	myargc		= init->myargc;

}

/**********************************************************************/

static void (*netget) (void);
static void (*netsend) (void);


/**********************************************************************

PacketSend:

Send the netbuffer contents (**netbuffer, because of netbuffer
being a pointer to the real variable) to the player
doomcom->remotenode. doomcom->remontenode is the ID of
the player. For example if DoomAttack was started with
"-net 1 John Peter" than:

 0 is you
 1 is John
 2 is Peter
 
You need to transfer only doomcom->datalength bytes.

**********************************************************************/

static void PacketSend (void)
{
}


/**********************************************************************

PacketGet:

Check if there's any new packet (data) from another
player. If there isn't set doomcom->remotenode to -1.

If there is, then set doomcom->remotnenode to the ID of
the player the packet comes from (see PacketSend).
Set also the length of the data: doomcom->datalength
and then copy the recieved data to the netbuffer
(*netbuffer, because of netbuffer being a pointer to
the real variable).


**********************************************************************/

static void PacketGet (void)
{
}

/**********************************************************************

DAN_InitNetwork:

Do our init stuff and return TRUE if everything is OK.
If something went wrong cleanup and return FALSE - in
this case DoomAttack won't call DAN_CleanupNetwork.

If you return TRUE you must set doomcom->numplayers
and doomcom->numnodes to the number of players.
Further you have to set doomcom->consoleplayer to
the id of the "locale" player. All player ids are
in the range between 0 and <num players-1>. For example
if DoomAttack was started with "-net 3 Joe" then you
would set doomcom->consoleplayer to 2! This is not
the same as in PacketSend/PacketGet!

I nearly forgot! You must also set doomcom->id to
DOOMCOM_ID. Doom uses this to check whether the
different players are using the same version of Doom
(different version will not be accepted).

**********************************************************************/

int DAN_InitNetwork (void)
{
  int i,i2;
  int netgame=0;

  char trueval = true;

  // set up for network
  i = M_CheckParm ("-dup");
  if (i && i < myargc - 1) {
    doomcom->ticdup = myargv[i+1][0]-'0';
    if (doomcom->ticdup < 1)
      doomcom->ticdup = 1;
    if (doomcom->ticdup > 9)
      doomcom->ticdup = 9;
  } else
    doomcom-> ticdup = 1;

  if (M_CheckParm ("-extratic"))
    doomcom-> extratics = 1;
  else
    doomcom-> extratics = 0;

  // parse network game options,
  //  -net <consoleplayer> <host> <host> ...

  // don't need to check if "-net" option is present.
  // DoomAttack does this for you and if it isn't
  // present, it does not call DAM_InitNetwork

  i = M_CheckParm ("-net");

  netsend = PacketSend;
  netget = PacketGet;

  netgame = 1;

  // parse player number and host list
  doomcom->consoleplayer = consoleplayer = myargv[i+1][0]-'1';

  doomcom->numnodes = 1; // this node for sure

  i++;
  while (++i < myargc && myargv[i][0] != '-')
  {
    doomcom->numnodes++;
  }

  doomcom->id = DOOMCOM_ID;
  doomcom->numplayers = doomcom->numnodes;

  ...

  return ????;
}


/**********************************************************************

DAN_NetCmd:

If doomcom->command == CMD_SEND call netsend()
If doomcom->command == CMD_GET call netget()

You don't need to change this

**********************************************************************/


void DAN_NetCmd (void)
{
  if (doomcom->command == CMD_SEND) {
    netsend ();
  } else if (doomcom->command == CMD_GET) {
    netget ();
  } else
    I_Error ("Bad net cmd: %i\n",doomcom->command);
}

/**********************************************************************

DAN_CleanupNetwork:

Called when the user quits DoomAttack. Cleanup everything.

**********************************************************************/


void DAN_CleanupNetwork (void)
{
#ifdef __MAXON__
	CleanupModules();
#endif

}

/**********************************************************************/

